# Copyright 2010-2012 by Carnegie Mellon University

# @OPENSOURCE_HEADER_START@
# Use of the Prism trend script and related source code is subject
# to the terms of the following licenses:
# 
# GNU Public License (GPL) Rights pursuant to Version 2, June 1991
# Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
# 
# NO WARRANTY
# 
# ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
# PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
# PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
# "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
# KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
# LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
# MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
# OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
# SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
# TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
# WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
# LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
# CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
# CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
# DELIVERABLES UNDER THIS LICENSE.
# 
# Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
# Mellon University, its trustees, officers, employees, and agents from 
# all claims or demands made against them (and any related losses, 
# expenses, or attorney's fees) arising out of, or relating to Licensee's 
# and/or its sub licensees' negligent use or willful misuse of or 
# negligent conduct or willful misconduct regarding the Software, 
# facilities, or other rights or assistance granted by Carnegie Mellon 
# University under this License, including, but not limited to, any 
# claims of product liability, personal injury, death, damage to 
# property, or violation of any laws or regulations.
# 
# Carnegie Mellon University Software Engineering Institute authored 
# documents are sponsored by the U.S. Department of Defense under 
# Contract FA8721-05-C-0003. Carnegie Mellon University retains 
# copyrights in all material produced under this contract. The U.S. 
# Government retains a non-exclusive, royalty-free license to publish or 
# reproduce these documents, or allow others to do so, for U.S. 
# Government purposes only pursuant to the copyright license under the 
# contract clause at 252.227.7013.
# @OPENSOURCE_HEADER_END@

from netsa.data.times import make_datetime
from netsa.util.shell import command

# This chart support is simply for color spec parsing
chart_support = False

try:
    from rayon.miscutils import InvalidColorSpec, parse_color_string
    import rayon.toolbox
    import rayon.backend_cairo
    chart_support = True
except:
    pass

from datetime import timedelta

########################################################################

_key_id = {}
_id_key = {}
_next_key_id = 0

def get_key_int_id(key):
    try:
        return _key_id[key]
    except KeyError:
        global _next_key_id
        _key_id[key] = _next_key_id
        _id_key[_next_key_id] = key
        _next_key_id += 1
        return _key_id[key]

def get_int_id_key(id):
    return _id_key[id]

########################################################################

class TrendConfigError(Exception):
    pass

########################################################################

class trend_bin(object):
    """
    A trend bin represents an individual bin in a trend scheme.  Note
    that the contents and meaning of a bin cannot be determined
    without knowing what comes before it in the trend scheme: any
    traffic accept by earlier bins won't ever be accepted for this
    one.  Each bin has a name, a set of rwfilter switches, and some
    number of rendering attributes.
    """
    __slots__ = ["_name", "_title", "_title_font_size",
                 "_in_filter_args", "_out_filter_args"]
    def __init__(self, name, title, title_font_size,
                 in_filter_args, out_filter_args):
        if title == None:
            title = name
        self._name = name
        self._title = title
        self._title_font_size = title_font_size
        self._in_filter_args = in_filter_args
        self._out_filter_args = out_filter_args
    def get_name(self):
        return self._name
    def get_title(self):
        return self._title
    def get_title_font_size(self):
        return self._title_font_size
    def get_in_filter_args(self):
        return self._in_filter_args
    def get_out_filter_args(self):
        return self._out_filter_args
    def get_in_command(self, filename):
        return command("rwfilter", "--input-pipe=stdin",
                       "--pass-destination=" + filename,
                       "--fail-destination=stdout",
                       "--compression-method=none",
                       *self._in_filter_args)
    def get_out_command(self, filename):
        return command("rwfilter", "--input-pipe=stdin",
                       "--pass-destination=" + filename,
                       "--fail-destination=stdout",
                       "--compression-method=none",
                       *self._out_filter_args)

class leftover_trend_bin(trend_bin):
    """
    A leftover trend bin is a bin for collecting all of the "leftovers"
    at the end of a trend scheme.
    """
    def __init__(self):
        trend_bin.__init__(self, "other", "Other", 14,
                           ["--protocol=0-255"], ["--protocol=0-255"])
    def get_in_command(self, filename):
        return command("rwfilter", "--input-pipe=stdin",
                       "--pass-destination=" + filename,
                       "--all-destination=stdout",
                       "--compression-method=none",
                       "--proto=0-255")
    def get_out_command(self, filename):
        return self.get_in_command(filename)

class trend_scheme(object):
    """
    A trend scheme is a sequence of trend bins.  When processing, each
    bin consumes some portion of the input and then only the portion
    of the input that is not consumed is passed on to the other bins.
    Hence, you need the whole sequence of bins leading up to a given
    bin to know what goes into it.
    """
    __slots__ = ["_bins", "_bin_by_label"]
    def __init__(self):
        self._bins = []
    def add_bin(self, bin):
        self._bins.append(bin)
    def _get_bin_titles(self):
        for bin in self._bins:
	    yield bin.get_title()
    def get_bin_titles(self):
        return list(self._get_bin_titles())
    def _get_bin_title_font_sizes(self):
        for bin in self._bins:
	    yield bin.get_title_font_size()
    def get_bin_title_font_sizes(self):
        return list(self._get_bin_title_font_sizes())
    def _get_bin_names(self):
        for bin in self._bins:
            yield bin.get_name()
    def get_bin_names(self):
        return list(self._get_bin_names())
    def _get_in_bin_keys(self):
        if self._bin_by_label:
            for bin in self._bins:
                yield get_key_int_id('bin:' + bin.get_name())
        else:
            result = []
            for bin in self._bins:
                result.append(' '.join(a.lstrip('-')
                                       for a in bin.get_in_filter_args()))
                yield get_key_int_id('/'.join(result))
    def _get_out_bin_keys(self):
        if self._bin_by_label:
            for bin in self._bins:
                yield get_key_int_id('bin:' + bin.get_name())
        else:
            result = []
            for bin in self._bins:
                result.append(' '.join(a.lstrip('-')
                                       for a in bin.get_out_filter_args()))
                yield get_key_int_id('/'.join(result))
    def bin_by_label(self, flag):
        self._bin_by_label = bool(flag)
    def get_in_bin_keys(self):
        return list(self._get_in_bin_keys())
    def get_out_bin_keys(self):
        return list(self._get_out_bin_keys())
    def get_bins(self):
        return list(self._bins)

########################################################################

def parse_config_line(config, l):
    l = l.strip()
    if l.startswith('#'):
        return
    if l == '':
        return
    split_line = l.split(':', 1)
    name = split_line[0].strip()
    value = None
    if len(split_line) == 2:
        value = split_line[1].strip()
    if name == 'include':
        parse_config_file(config, value)
    elif name == 'database-uri':
        config['database_uri'] = value
    elif name == 'database-store-by-label':
        value = value.lower()
        if value in ('1', 'true', 't', 'yes', 'y', 'on'):
            config['database_store_by_label'] = True
        else:
            config['database_store_by_label'] = False
    elif name == 'in-rwfilter':
        config['in_rwfilter'] = value
    elif name == 'out-rwfilter':
        config['out_rwfilter'] = value
    elif name == 'in-types':
        config['in_types'] = value
    elif name == 'out-types':
        config['out_types'] = value
    elif name == 'chart-out-color':
        if chart_support:
            try:
                config['chart_out_color'] = parse_color_string(value)
            except InvalidColorSpec:
                raise TrendConfigError(
                    "config: chart-out-color %s is not a valid color" %
                    repr(value))
    elif name == 'chart-in-color':
        if chart_support:
            try:
                config['chart_in_color'] = parse_color_string(value)
            except InvalidColorSpec:
                raise TrendConfigError(
                    "config: chart-in-color %s is not a valid color" %
                    repr(value))
    elif name == 'chart-title':
        config['chart_title'] = value
    elif name == 'chart-title':
        config['chart_title'] = value
    elif name == 'chart-padding-vertical':
        try:
            config['chart_padding_vertical'] = int(value)
        except ValueError:
            raise TrendConfigError(
                "config: chart-padding-vertical %s must be integer" %
                repr(value))
    elif name == 'chart-padding-horizontal':
        try:
            config['chart_padding_horizontal'] = int(value)
        except ValueError:
            raise TrendConfigError(
                "config: chart-padding-horizontal %s must be integer" %
                repr(value))
    elif name == 'chart-padding':
        try:
            config['chart_padding'] = int(value)
        except ValueError:
            raise TrendConfigError(
                "config: chart-padding %s must be integer" %
                repr(value))
    elif name == 'chart-strip-height':
        try:
            config['chart_strip_height'] = int(value)
        except ValueError:
            raise TrendConfigError(
                "config: chart-strip-height %s must be integer" %
                repr(value))
    elif name == 'chart-height':
        try:
            config['chart_height'] = int(value)
        except ValueError:
            raise TrendConfigError(
                "config: chart-height %s must be integer" %
                repr(value))
    elif name == 'chart-width':
        try:
            config['chart_width'] = int(value)
        except ValueError:
            raise TrendConfigError(
                "config: chart-width %s must be integer" %
                repr(value))
    elif name == 'invertible-filters':
        for p in value.split(','):
            ab = [v.strip() for v in p.split('/')]
            if len(ab) != 2:
                raise TrendConfigError(
                    "config: invertible-filter %r must contain exactly one '/'"
                    % p)
            (a, b) = ab
            if not a.startswith('--'):
                a = '--' + a
            if not b.startswith('--'):
                b = '--' + b
            for x in (a, b):
                if x in config['invertible_filters']:
                    y = config['invertible_filters'][x]
                    del config['invertible_filters'][y]
                    del config['invertible_filters'][x]
            config['invertible_filters'][a] = b
            config['invertible_filters'][b] = a
    elif name == '[bin]':
        if config['_bin_info'] != None:
            # Handle old bin
            config['trend_scheme'].add_bin(
                trend_bin(config['_bin_info']['name'],
                          config['_bin_info']['title'],
                          config['_bin_info']['title_font_size'],
                          config['_bin_info']['in_filters'],
                          config['_bin_info']['out_filters']))
        config['_bin_info'] = {
            'name': '',
            'in_filters': [],
            'out_filters': [],
            'title': None,
            'title_font_size': 14,
            }
    elif name == 'filter':
        if not value.startswith('--'):
            value = '--' + value
        config['_bin_info']['in_filters'].append(value)
        config['_bin_info']['out_filters'].append(value)
    elif name == 'in-out-filter':
        if not value.startswith('--'):
            value = '--' + value
        if '=' in value:
            (switch, arg) = value.split('=', 1)
        else:
            switch = value
            arg = None
        inv_switch = config['invertible_filters'].get(switch, None)
        if inv_switch == None:
            raise TrendConfigError(
                "config: could not find inverse for in-out-filter %r" % value)
        if arg == None:
            inv_value = inv_switch
        else:
            inv_value = inv_switch + '=' + arg
        config['_bin_info']['in_filters'].append(value)
        config['_bin_info']['out_filters'].append(inv_value)
    elif name == 'out-in-filter':
        if not value.startswith('--'):
            value = '--' + value
        if '=' in value:
            (switch, arg) = value.split('=', 1)
        else:
            switch = value
            arg = None
        inv_switch = config['invertible_filters'].get(switch, None)
        if inv_switch == None:
            raise TrendConfigError(
                "config: could not find inverse for out-in-filter %r" % value)
        if arg == None:
            inv_value = inv_switch
        else:
            inv_value = inv_switch + '=' + arg
        config['_bin_info']['out_filters'].append(value)
        config['_bin_info']['in_filters'].append(inv_value)
    elif name == 'in-filter':
        if not value.startswith('--'):
            value = '--' + value
        config['_bin_info']['in_filters'].append(value)
    elif name == 'out-filter':
        if not value.startswith('--'):
            value = '--' + value
        config['_bin_info']['out_filters'].append(value)
    elif name == 'name':
        config['_bin_info']['name'] = value
    elif name == 'title':
        config['_bin_info']['title'] = value
    elif name == 'title-font-size':
        try:
            config['_bin_info']['title_font_size'] = int(value)
        except ValueError:
            raise TrendConfigError(
                "config: title-font-size %s must be integer" %
                repr(value))
    else:
        raise TrendConfigError(
            "Unrecognized config option: %s" % repr(name))

def parse_config_file(config, filename):
    f = None
    try:
        f = open(filename, 'r')
        for l in f:
            parse_config_line(config, l)
    finally:
        if f != None:
            f.close()

def read_config(filename):
    config = {}
    config['trend_scheme'] = trend_scheme()
    config['database_uri'] = None
    config['database_store_by_label'] = False
    config['in_rwfilter'] = None
    config['out_rwfilter'] = None
    config['in_types'] = None
    config['out_types'] = None
    config['chart_out_color'] = (255,0,0,128)
    config['chart_in_color'] = (0,0,255,128)
    config['invertible_filters'] = {}
    config['_bin_info'] = None
    parse_config_file(config, filename)
    if config['_bin_info'] != None:
        config['trend_scheme'].add_bin(
            trend_bin(config['_bin_info']['name'],
                      config['_bin_info']['title'],
                      config['_bin_info']['title_font_size'],
                      config['_bin_info']['in_filters'],
                      config['_bin_info']['out_filters']))
    config['trend_scheme'].add_bin(leftover_trend_bin())
    config['trend_scheme'].bin_by_label(config['database_store_by_label'])
    return config

########################################################################

def gen_time_ticks(col):
    import calendar
    def as_seconds(d):
        return calendar.timegm(d.timetuple())
    time_min = col.min()
    time_max = col.max()
    
    if (time_max - time_min) < timedelta(hours=25):
        tick_data = [make_datetime(s)
                     for s in xrange(int(as_seconds(time_min)),
                                     int(as_seconds(time_max)), 3600)]
        tick_data_minor = [d for d in tick_data if d.hour % 4 != 0]
        tick_data_major = [d for d in tick_data if d.hour % 4 == 0]
        tick_data_labeler = (lambda x: "%02d:%02d" % (x.hour, x.minute))
    elif (time_max - time_min) < timedelta(days=15):
        tick_data = [make_datetime(s)
                     for s in xrange(int(as_seconds(time_min)),
                                     int(as_seconds(time_max)), 4*3600)]
        tick_data_minor = [d for d in tick_data if d.hour != 0]
        tick_data_major = [d for d in tick_data if d.hour == 0]
        tick_data_labeler = (lambda x: "%02d/%02d" % (x.month, x.day))
    elif (time_max - time_min) < timedelta(days=61):
        tick_data = [make_datetime(s)
                     for s in xrange(int(as_seconds(time_min)),
                                     int(as_seconds(time_max)), 86400)]
        tick_data_minor = [d for d in tick_data if d.weekday() != 6]
        tick_data_major = [d for d in tick_data if d.weekday() == 6]
        tick_data_labeler = (lambda x: "%02d/%02d" % (x.month, x.day))
    else:
        tick_data = [make_datetime(s)
                     for s in xrange(int(as_seconds(time_min)),
                                     int(as_seconds(time_max)), 86400)]
        tick_data_minor = [d for d in tick_data if d.weekday() == 6]
        tick_data_major = [d for d in tick_data if d.day == 1]
        tick_data_labeler = (lambda x: "%04d/%02d" % (x.year, x.month))
    return (tick_data_minor, tick_data_major, tick_data_labeler)

# dataset has columns "time", "in", and "out"
def make_chart(toolbox, time_scale, dataset, title, title_font_size, anno_func,
               background_color, color_out, color_in):

    y_max_val = max(dataset.column("out").max(), dataset.column("in").max())

    y_scale_out = toolbox.new_scale("linear",
                                    a_lo=0.0, a_hi=y_max_val,
                                    b_lo=0.5, b_hi=1.0)
    y_scale_in = toolbox.new_scale("linear",
                                   a_lo=0.0, a_hi=y_max_val,
                                   b_lo=0.5, b_hi=0.0)

    pt_out_anno = []
    for r in dataset:
        if r['out'] == dataset.column('out').max():
            p = [r['time'], r['out']]
            pt_out_anno.append(p + [anno_func(p)])
            break
    dataset_out_anno = rayon.data.Dataset.from_rows(
        pt_out_anno, colnames=["time", "out", "text"])

    pt_in_anno = []
    for r in dataset:
        if r['in'] == dataset.column('in').max():
            p = [r['time'], r['in']]
            pt_in_anno.append(p + [anno_func(p)])
            break
    dataset_in_anno = rayon.data.Dataset.from_rows(
        pt_in_anno, colnames=["time", "in", "text"])


    chart = toolbox.new_chart("square")

    plot_out = toolbox.new_plot(
        "filledline", fill_color=color_out,
        x_data=dataset.column('time'), y_data=dataset.column('out'),
        x_scale=time_scale, y_scale=y_scale_out,
        line_color_scale="black", line_width_scale=1,
        line_style_scale="solid")
    chart.add_plot(plot_out)

    plot_in = toolbox.new_plot(
        "filledline", fill_color=color_in,
        x_data=dataset.column('time'), y_data=dataset.column('in'),
        x_scale=time_scale, y_scale=y_scale_in,
        line_color_scale="black", line_width_scale=1,
        line_style_scale="solid")
    chart.add_plot(plot_in)

    plot_out_anno = toolbox.new_plot(
        "labeledscatter",
        x_data=dataset_out_anno.column("time"),
        y_data=dataset_out_anno.column("out"),
        text_data=dataset_out_anno.column("text"),
        x_scale=time_scale, y_scale=y_scale_out,
        marker_color_scale="black", marker_shape_scale="circle",
        marker_size_scale=5, text_scale=lambda s: s,
        label_position_scale="ne", label_spacing_scale=5,
        font_properties_scale={'halign': 'left', 'valign': 'center',
                               'font_size': 10},
        border_properties_scale={})
    plot_out_anno.set_options(avoid_horizontal_overflow=True)
    chart.add_plot(plot_out_anno)

    plot_in_anno = toolbox.new_plot(
        "labeledscatter",
        x_data=dataset_in_anno.column("time"),
        y_data=dataset_in_anno.column("in"),
        text_data=dataset_in_anno.column("text"),
        x_scale=time_scale, y_scale=y_scale_in,
        marker_color_scale="black", marker_shape_scale="circle",
        marker_size_scale=5, text_scale=lambda s: s,
        label_position_scale="se", label_spacing_scale=5,
        font_properties_scale={'halign': 'left', 'valign': 'center',
                               'font_size': 10},
        border_properties_scale={})    
    plot_in_anno.set_options(avoid_horizontal_overflow=True)
    chart.add_plot(plot_in_anno)

    border_line = toolbox.new_border(
        'vline', line=toolbox.new_line('solid', color=(0,0,0), width=2))
    chart.add_left_border(border_line, "2px")

    border_label_out = toolbox.new_border(
        'vlabel', label='OUT', halign="center", valign="bottom",
        font_size=10)
    border_label_in = toolbox.new_border(
        'vlabel', label='IN', halign="center", valign="bottom",
        font_size=10)
    border_inout = toolbox.new_border(
        'vsplit', children=[border_label_out, border_label_in], rpad="1px")
    chart.add_left_border(border_inout, width="12px")
    
    border_title = toolbox.new_border(
        'hlabel', label=title, font_weight="bold", font_size=title_font_size,
        halign="right", valign="center")
    chart.add_left_border(border_title, width="125px")

    (tick_data_minor, tick_data_major, tick_major_labeler) = \
        gen_time_ticks(dataset.column("time"))

    tick_labeler = toolbox.new_label(font_size=10)

    minor_tickmarker = toolbox.new_tickmarker(
        "horizontal", marker_args={'above': 2, 'below': 2},
        labeler=tick_labeler)
    minor_tickset = toolbox.new_tickset(
        "default", tick_tuples=[(d, "") for d in tick_data_minor],
        scale=time_scale, marker=minor_tickmarker)

    major_tickmarker = toolbox.new_tickmarker(
        "horizontal", marker_args={'above': 4, 'below': 4},
        labeler=tick_labeler)
    major_tickset = toolbox.new_tickset(
        "default", tick_tuples=[(d, "") for d in tick_data_major],
        scale=time_scale, marker=major_tickmarker)

    time_axis = toolbox.new_gridlines("horizontal")
    time_axis.set_data(
        pos=rayon.data.Column([0]),
        ticksets=rayon.data.Column([[minor_tickset, major_tickset]]))
    time_axis.set_scales(
        pos=y_scale_out,
        ticksets=lambda x: x,
        line_style='solid',
        line_width=1,
        line_color='black')
    
    y_grid = toolbox.new_gridlines("vertical")
    y_grid.set_data(
        pos=rayon.data.Column(tick_data_major))
    y_grid.set_scales(
        pos=time_scale,
        ticksets=lambda x: None,
        line_style="dotted",
        line_width=1,
        line_color=(32, 32, 32))

    chart.add_front_gridlines(time_axis)
    chart.add_rear_gridlines(y_grid)
    chart.set_chart_background(background_color)

    return chart

def write_chart_single(out_file, out_type, width, height,
                       time_data, out_data, in_data,
                       title, title_font_size, anno_func, 
                       background_color="rgba(255,255,255)",
                       out_color="rgba(255,0,0,0.5)",
                       in_color="rgba(0,0,255,0.5)"):

    toolbox = rayon.toolbox.Toolbox.for_backend("cairo")

    dataset = rayon.data.Dataset.from_columns(
        [time_data, out_data, in_data],
        header={'title': title},
        colnames=['time', 'out', 'in'])

    time_scale = toolbox.new_scale(
        "time",
        begin=dataset.column("time").min(),
        end=dataset.column("time").max())

    (tick_data_minor, tick_data_major, tick_major_labeler) = \
        gen_time_ticks(dataset.column("time"))

    tick_labeler = toolbox.new_label(font_size=10)

    major_tickmarker = toolbox.new_tickmarker(
        "horizontal", marker_args={'above': 4, 'below': 4},
        labeler=tick_labeler)
    major_tickset = toolbox.new_tickset(
        "default", tick_tuples=[(d, tick_major_labeler(d))
                                for d in tick_data_major],
        scale=time_scale, marker=major_tickmarker)

    chart = make_chart(toolbox, time_scale, dataset, title, title_font_size,
                       anno_func, background_color, out_color, in_color)

    chart.set_padding(tpad="15px", lpad="15px", rpad="15px")

    border = toolbox.new_border(
        "hline", line=toolbox.new_line("solid", width=2), tpad="8px")
    # Padding is actually 16px, tpad is doubled
    border.add_tickset(major_tickset)
    chart.add_bottom_border(border, height="36px")

    page = rayon.backend_cairo.Page.from_file_object(
        out_file, out_type, width, height)
    page.write(chart)

def write_chart_multi(out_file, out_type, width, height,
                      time_data, out_datas, in_datas,
                      titles, title_font_sizes,
                      anno_func, chart_title="",
                      background_color="rgb(255,255,255)",
                      out_color="rgba(255,0,0,0.5)",
                      in_color="rgba(0,0,255,0.5)"):

    toolbox = rayon.toolbox.Toolbox.for_backend("cairo")

    time_col = rayon.data.Column(time_data)

    time_scale = toolbox.new_scale(
        "time", begin=time_col.min(), end=time_col.max())

    multichart = toolbox.new_chart("tiled", num_columns=1,
                                   orientation="vertical")

    (tick_data_minor, tick_data_major, tick_major_labeler) = \
        gen_time_ticks(time_col)

    tick_labeler = toolbox.new_label(font_size=10)

    major_tickmarker = toolbox.new_tickmarker(
        "horizontal", marker_args={'above': 4, 'below': 4},
        labeler=tick_labeler)
    major_tickset = toolbox.new_tickset(
        "default", tick_tuples=[(d, tick_major_labeler(d))
                                for d in tick_data_major],
        scale=time_scale, marker=major_tickmarker)

    for out_data, in_data, title, title_font_size \
            in zip(out_datas, in_datas, titles, title_font_sizes):
        dataset = rayon.data.Dataset.from_columns(
            [time_data, out_data, in_data],
            header={'title': title},
            colnames=['time', 'out', 'in'])
        chart = make_chart(toolbox, time_scale, dataset, title,
                           title_font_size, anno_func,
                           background_color, out_color, in_color)
        chart.set_padding(tpad="10px", bpad="10px", lpad="15px", rpad="15px")
        multichart.add_chart(chart)

    multichart.set_padding(tpad="10px", bpad="10px")
    multichart.set_chart_background(background_color)

    if chart_title:
        top_title = toolbox.new_border('hlabel', label=chart_title,
                                       font_size="18px", valign="top",
                                       halign="center")
        multichart.add_top_border(top_title, height="24px")

    if chart:
        border = toolbox.new_border(
            "hline", line=toolbox.new_line("solid", width=2), tpad="8px")
        border.add_tickset(major_tickset)
        chart.add_bottom_border(border, height="36px")

    page = rayon.backend_cairo.Page.from_file_object(
        out_file, out_type, width, height)
    page.write(multichart)

__all__ = """

    TrendConfigError

    get_key_int_id
    get_int_id_key

    read_config

    write_chart_single
    write_chart_multi

""".split()
